'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var neoBlockchainCore = require('neo-blockchain-core');
var neoBlockchainNodeCore = require('neo-blockchain-node-core');
var PriorityQueue = _interopDefault(require('js-priority-queue'));
var _ = _interopDefault(require('lodash'));
var perf_hooks = require('perf_hooks');
var rxjs = require('rxjs');

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          Promise.resolve(value).then(_next, _throw);
        }
      }

      function _next(value) {
        step("next", value);
      }

      function _throw(err) {
        step("throw", err);
      }

      _next();
    });
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _sliceIterator(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _slicedToArray(arr, i) {
  if (Array.isArray(arr)) {
    return arr;
  } else if (Symbol.iterator in Object(arr)) {
    return _sliceIterator(arr, i);
  } else {
    throw new TypeError("Invalid attempt to destructure non-iterable instance");
  }
}

function _toConsumableArray(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
}

var _fixBabelExtend = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var GenesisBlockNotRegisteredError = _fixBabelExtend(
/*#__PURE__*/
function (_Error) {
  _inherits(GenesisBlockNotRegisteredError, _Error);

  function GenesisBlockNotRegisteredError() {
    _classCallCheck(this, GenesisBlockNotRegisteredError);
    return _possibleConstructorReturn(this, (GenesisBlockNotRegisteredError.__proto__ || Object.getPrototypeOf(GenesisBlockNotRegisteredError)).call(this, 'Genesis block was not registered with storage.'));
  }

  return GenesisBlockNotRegisteredError;
}(Error));
var VerifyError = _fixBabelExtend(
/*#__PURE__*/
function (_Error2) {
  _inherits(VerifyError, _Error2);

  function VerifyError() {
    _classCallCheck(this, VerifyError);
    return _possibleConstructorReturn(this, (VerifyError.__proto__ || Object.getPrototypeOf(VerifyError)).call(this, 'Script verification failed.'));
  }

  return VerifyError;
}(Error));
var WitnessVerifyError = _fixBabelExtend(
/*#__PURE__*/
function (_Error3) {
  _inherits(WitnessVerifyError, _Error3);

  function WitnessVerifyError() {
    _classCallCheck(this, WitnessVerifyError);
    return _possibleConstructorReturn(this, (WitnessVerifyError.__proto__ || Object.getPrototypeOf(WitnessVerifyError)).call(this, 'Witness verification failed.'));
  }

  return WitnessVerifyError;
}(Error));

function createGet(_ref) {
  var tryGetTracked = _ref.tryGetTracked,
      readStorage = _ref.readStorage;
  return function () {
    var _ref2 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee(key) {
      var trackedChange, value;
      return regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              trackedChange = tryGetTracked(key);

              if (!(trackedChange != null)) {
                _context.next = 5;
                break;
              }

              if (!(trackedChange.type === 'delete')) {
                _context.next = 4;
                break;
              }

              throw new Error('Not found');

            case 4:
              return _context.abrupt("return", trackedChange.value);

            case 5:
              _context.next = 7;
              return readStorage.get(key);

            case 7:
              value = _context.sent;
              return _context.abrupt("return", value);

            case 9:
            case "end":
              return _context.stop();
          }
        }
      }, _callee, this);
    }));

    return function (_x) {
      return _ref2.apply(this, arguments);
    };
  }();
}

function createTryGet(_ref3) {
  var tryGetTracked = _ref3.tryGetTracked,
      readStorage = _ref3.readStorage;
  return function () {
    var _ref4 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee2(key) {
      var trackedChange, value;
      return regeneratorRuntime.wrap(function _callee2$(_context2) {
        while (1) {
          switch (_context2.prev = _context2.next) {
            case 0:
              trackedChange = tryGetTracked(key);

              if (!(trackedChange != null)) {
                _context2.next = 5;
                break;
              }

              if (!(trackedChange.type === 'delete')) {
                _context2.next = 4;
                break;
              }

              return _context2.abrupt("return", null);

            case 4:
              return _context2.abrupt("return", trackedChange.value);

            case 5:
              _context2.next = 7;
              return readStorage.tryGet(key);

            case 7:
              value = _context2.sent;
              return _context2.abrupt("return", value);

            case 9:
            case "end":
              return _context2.stop();
          }
        }
      }, _callee2, this);
    }));

    return function (_x2) {
      return _ref4.apply(this, arguments);
    };
  }();
}

var BaseReadStorageCache =
/*#__PURE__*/
function () {
  function BaseReadStorageCache(options) {
    _classCallCheck(this, BaseReadStorageCache);
    this._readStorage = options.readStorage;
    this._name = options.name;
    this._createAddChange = options.createAddChange;
    this._createDeleteChange = options.createDeleteChange;
    this._values = {};
    this.get = createGet({
      readStorage: this._readStorage,
      tryGetTracked: this._tryGetTracked.bind(this)
    });
    this.tryGet = createTryGet({
      readStorage: this._readStorage,
      tryGetTracked: this._tryGetTracked.bind(this)
    });
  }

  _createClass(BaseReadStorageCache, [{
    key: "getChangeSet",
    value: function getChangeSet() {
      var _this = this;

      var createDeleteChange = this._createDeleteChange;
      return neoBlockchainCore.utils.values(this._values).map(function (value) {
        if (value.type === 'delete') {
          if (createDeleteChange == null) {
            // TODO: Make better
            throw new Error('Invalid delete');
          }

          return {
            type: 'delete',
            change: createDeleteChange(value.key)
          };
        }

        return {
          type: 'add',
          change: _this._createAddChange(value.value)
        };
      });
    } // eslint-disable-next-line

  }, {
    key: "_tryGetTracked",
    value: function _tryGetTracked(key) {
      throw new Error('Not Implemented');
    }
  }]);
  return BaseReadStorageCache;
}();

var ReadStorageCache =
/*#__PURE__*/
function (_BaseReadStorageCache) {
  _inherits(ReadStorageCache, _BaseReadStorageCache);

  function ReadStorageCache(options) {
    var _this2;

    _classCallCheck(this, ReadStorageCache);
    _this2 = _possibleConstructorReturn(this, (ReadStorageCache.__proto__ || Object.getPrototypeOf(ReadStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this2._getKeyString = options.getKeyString;
    return _this2;
  }

  _createClass(ReadStorageCache, [{
    key: "_tryGetTracked",
    value: function _tryGetTracked(key) {
      return this._values[this._getKeyString(key)];
    }
  }]);
  return ReadStorageCache;
}(BaseReadStorageCache);

var ReadAllStorageCache =
/*#__PURE__*/
function (_ReadStorageCache) {
  _inherits(ReadAllStorageCache, _ReadStorageCache);

  function ReadAllStorageCache(options) {
    var _this3;

    _classCallCheck(this, ReadAllStorageCache);
    _this3 = _possibleConstructorReturn(this, (ReadAllStorageCache.__proto__ || Object.getPrototypeOf(ReadAllStorageCache)).call(this, {
      readStorage: {
        get: options.readAllStorage.get,
        tryGet: options.readAllStorage.tryGet
      },
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this3._readAllStorage = options.readAllStorage;
    _this3._getKeyFromValue = options.getKeyFromValue;
    _this3.all = rxjs.Observable.concat(rxjs.Observable.defer(function () {
      return rxjs.Observable.of.apply(rxjs.Observable, _toConsumableArray(neoBlockchainCore.utils.values(_this3._values).map(function (value) {
        return value.type === 'add' ? value.value : null;
      }).filter(Boolean)));
    }), _this3._readAllStorage.all.concatMap(function (value) {
      var trackedChange = _this3._tryGetTracked(_this3._getKeyFromValue(value));

      if (trackedChange != null && trackedChange.type === 'delete') {
        return rxjs.Observable.of();
      }

      return rxjs.Observable.of(value);
    }));
    return _this3;
  }

  return ReadAllStorageCache;
}(ReadStorageCache);

var ReadGetAllStorageCache =
/*#__PURE__*/
function (_ReadStorageCache2) {
  _inherits(ReadGetAllStorageCache, _ReadStorageCache2);

  function ReadGetAllStorageCache(options) {
    var _this4;

    _classCallCheck(this, ReadGetAllStorageCache);
    _this4 = _possibleConstructorReturn(this, (ReadGetAllStorageCache.__proto__ || Object.getPrototypeOf(ReadGetAllStorageCache)).call(this, {
      readStorage: {
        get: options.readGetAllStorage.get,
        tryGet: options.readGetAllStorage.tryGet
      },
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this4._readGetAllStorage = options.readGetAllStorage;
    _this4._getKeyFromValue = options.getKeyFromValue;
    _this4._matchesPartialKey = options.matchesPartialKey;

    _this4.getAll = function (key) {
      var createdValues = neoBlockchainCore.utils.values(_this4._values).map(function (value) {
        return value.type === 'add' && _this4._matchesPartialKey(value.value, key) ? value.value : null;
      }).filter(Boolean);
      return rxjs.Observable.concat(rxjs.Observable.of.apply(rxjs.Observable, _toConsumableArray(createdValues)), _this4._readGetAllStorage.getAll(key).concatMap(function (value) {
        var trackedChange = _this4._tryGetTracked(_this4._getKeyFromValue(value));

        if (trackedChange != null && trackedChange.type === 'delete') {
          return rxjs.Observable.of();
        }

        return rxjs.Observable.of(value);
      }));
    };

    return _this4;
  }

  return ReadGetAllStorageCache;
}(ReadStorageCache);

function createAdd(_ref5) {
  var cache = _ref5.cache,
      getKeyFromValue = _ref5.getKeyFromValue,
      getKeyString = _ref5.getKeyString;
  return function () {
    var _ref6 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee3(value, force) {
      var key, currentValue;
      return regeneratorRuntime.wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              key = getKeyFromValue(value);

              if (force) {
                _context3.next = 7;
                break;
              }

              _context3.next = 4;
              return cache.tryGet(key);

            case 4:
              currentValue = _context3.sent;

              if (!(currentValue != null)) {
                _context3.next = 7;
                break;
              }

              throw new Error("Attempted to add an already existing object for key " + "".concat(cache._name, ":").concat(getKeyString(key), "."));

            case 7:
              // eslint-disable-next-line
              cache._values[cache._getKeyString(key)] = {
                type: 'add',
                value: value
              };

            case 8:
            case "end":
              return _context3.stop();
          }
        }
      }, _callee3, this);
    }));

    return function (_x3, _x4) {
      return _ref6.apply(this, arguments);
    };
  }();
}

function createUpdate(_ref7) {
  var cache = _ref7.cache,
      updateFunc = _ref7.update,
      getKeyFromValue = _ref7.getKeyFromValue;
  return function () {
    var _ref8 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee4(value, update) {
      var key, updatedValue;
      return regeneratorRuntime.wrap(function _callee4$(_context4) {
        while (1) {
          switch (_context4.prev = _context4.next) {
            case 0:
              key = getKeyFromValue(value);
              updatedValue = updateFunc(value, update); // eslint-disable-next-line

              cache._values[cache._getKeyString(key)] = {
                type: 'add',
                value: updatedValue
              };
              return _context4.abrupt("return", updatedValue);

            case 4:
            case "end":
              return _context4.stop();
          }
        }
      }, _callee4, this);
    }));

    return function (_x5, _x6) {
      return _ref8.apply(this, arguments);
    };
  }();
}

function createDelete(_ref9) {
  var cache = _ref9.cache;
  return function () {
    var _ref10 = _asyncToGenerator(
    /*#__PURE__*/
    regeneratorRuntime.mark(function _callee5(key) {
      return regeneratorRuntime.wrap(function _callee5$(_context5) {
        while (1) {
          switch (_context5.prev = _context5.next) {
            case 0:
              // eslint-disable-next-line
              cache._values[cache._getKeyString(key)] = {
                type: 'delete',
                key: key
              };

            case 1:
            case "end":
              return _context5.stop();
          }
        }
      }, _callee5, this);
    }));

    return function (_x7) {
      return _ref10.apply(this, arguments);
    };
  }();
}

var ReadAddUpdateDeleteStorageCache =
/*#__PURE__*/
function (_ReadStorageCache3) {
  _inherits(ReadAddUpdateDeleteStorageCache, _ReadStorageCache3);

  function ReadAddUpdateDeleteStorageCache(options) {
    var _this5;

    _classCallCheck(this, ReadAddUpdateDeleteStorageCache);
    _this5 = _possibleConstructorReturn(this, (ReadAddUpdateDeleteStorageCache.__proto__ || Object.getPrototypeOf(ReadAddUpdateDeleteStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this5.add = createAdd({
      cache: _this5,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    _this5.update = createUpdate({
      cache: _this5,
      update: options.update,
      getKeyFromValue: options.getKeyFromValue
    });
    _this5.delete = createDelete({
      cache: _this5
    });
    return _this5;
  }

  return ReadAddUpdateDeleteStorageCache;
}(ReadStorageCache);
var ReadAddUpdateStorageCache =
/*#__PURE__*/
function (_ReadStorageCache4) {
  _inherits(ReadAddUpdateStorageCache, _ReadStorageCache4);

  function ReadAddUpdateStorageCache(options) {
    var _this6;

    _classCallCheck(this, ReadAddUpdateStorageCache);
    _this6 = _possibleConstructorReturn(this, (ReadAddUpdateStorageCache.__proto__ || Object.getPrototypeOf(ReadAddUpdateStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this6.add = createAdd({
      cache: _this6,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    _this6.update = createUpdate({
      cache: _this6,
      update: options.update,
      getKeyFromValue: options.getKeyFromValue
    });
    return _this6;
  }

  return ReadAddUpdateStorageCache;
}(ReadStorageCache);
var ReadAddDeleteStorageCache =
/*#__PURE__*/
function (_ReadStorageCache5) {
  _inherits(ReadAddDeleteStorageCache, _ReadStorageCache5);

  function ReadAddDeleteStorageCache(options) {
    var _this7;

    _classCallCheck(this, ReadAddDeleteStorageCache);
    _this7 = _possibleConstructorReturn(this, (ReadAddDeleteStorageCache.__proto__ || Object.getPrototypeOf(ReadAddDeleteStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this7.add = createAdd({
      cache: _this7,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    _this7.delete = createDelete({
      cache: _this7
    });
    return _this7;
  }

  return ReadAddDeleteStorageCache;
}(ReadStorageCache);
var ReadAddStorageCache =
/*#__PURE__*/
function (_ReadStorageCache6) {
  _inherits(ReadAddStorageCache, _ReadStorageCache6);

  function ReadAddStorageCache(options) {
    var _this8;

    _classCallCheck(this, ReadAddStorageCache);
    _this8 = _possibleConstructorReturn(this, (ReadAddStorageCache.__proto__ || Object.getPrototypeOf(ReadAddStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange
    }));
    _this8.add = createAdd({
      cache: _this8,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    return _this8;
  }

  return ReadAddStorageCache;
}(ReadStorageCache);
var ReadGetAllAddUpdateDeleteStorageCache =
/*#__PURE__*/
function (_ReadGetAllStorageCac) {
  _inherits(ReadGetAllAddUpdateDeleteStorageCache, _ReadGetAllStorageCac);

  function ReadGetAllAddUpdateDeleteStorageCache(options) {
    var _this9;

    _classCallCheck(this, ReadGetAllAddUpdateDeleteStorageCache);
    _this9 = _possibleConstructorReturn(this, (ReadGetAllAddUpdateDeleteStorageCache.__proto__ || Object.getPrototypeOf(ReadGetAllAddUpdateDeleteStorageCache)).call(this, {
      readGetAllStorage: options.readGetAllStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange,
      getKeyFromValue: options.getKeyFromValue,
      matchesPartialKey: options.matchesPartialKey
    }));
    _this9.add = createAdd({
      cache: _this9,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    _this9.update = createUpdate({
      cache: _this9,
      update: options.update,
      getKeyFromValue: options.getKeyFromValue
    });
    _this9.delete = createDelete({
      cache: _this9
    });
    return _this9;
  }

  return ReadGetAllAddUpdateDeleteStorageCache;
}(ReadGetAllStorageCache);
var ReadGetAllAddStorageCache =
/*#__PURE__*/
function (_ReadGetAllStorageCac2) {
  _inherits(ReadGetAllAddStorageCache, _ReadGetAllStorageCac2);

  function ReadGetAllAddStorageCache(options) {
    var _this10;

    _classCallCheck(this, ReadGetAllAddStorageCache);
    _this10 = _possibleConstructorReturn(this, (ReadGetAllAddStorageCache.__proto__ || Object.getPrototypeOf(ReadGetAllAddStorageCache)).call(this, {
      readGetAllStorage: options.readGetAllStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange,
      getKeyFromValue: options.getKeyFromValue,
      matchesPartialKey: options.matchesPartialKey
    }));
    _this10.add = createAdd({
      cache: _this10,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    return _this10;
  }

  return ReadGetAllAddStorageCache;
}(ReadGetAllStorageCache);
var ReadAllAddUpdateDeleteStorageCache =
/*#__PURE__*/
function (_ReadAllStorageCache) {
  _inherits(ReadAllAddUpdateDeleteStorageCache, _ReadAllStorageCache);

  function ReadAllAddUpdateDeleteStorageCache(options) {
    var _this11;

    _classCallCheck(this, ReadAllAddUpdateDeleteStorageCache);
    _this11 = _possibleConstructorReturn(this, (ReadAllAddUpdateDeleteStorageCache.__proto__ || Object.getPrototypeOf(ReadAllAddUpdateDeleteStorageCache)).call(this, {
      readAllStorage: options.readAllStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange,
      getKeyFromValue: options.getKeyFromValue
    }));
    _this11.add = createAdd({
      cache: _this11,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    _this11.update = createUpdate({
      cache: _this11,
      update: options.update,
      getKeyFromValue: options.getKeyFromValue
    });
    _this11.delete = createDelete({
      cache: _this11
    });
    return _this11;
  }

  return ReadAllAddUpdateDeleteStorageCache;
}(ReadAllStorageCache);
var ReadAllAddStorageCache =
/*#__PURE__*/
function (_ReadAllStorageCache2) {
  _inherits(ReadAllAddStorageCache, _ReadAllStorageCache2);

  function ReadAllAddStorageCache(options) {
    var _this12;

    _classCallCheck(this, ReadAllAddStorageCache);
    _this12 = _possibleConstructorReturn(this, (ReadAllAddStorageCache.__proto__ || Object.getPrototypeOf(ReadAllAddStorageCache)).call(this, {
      readAllStorage: options.readAllStorage,
      name: options.name,
      getKeyString: options.getKeyString,
      createAddChange: options.createAddChange,
      createDeleteChange: options.createDeleteChange,
      getKeyFromValue: options.getKeyFromValue
    }));
    _this12.add = createAdd({
      cache: _this12,
      getKeyFromValue: options.getKeyFromValue,
      getKeyString: options.getKeyString
    });
    return _this12;
  }

  return ReadAllAddStorageCache;
}(ReadAllStorageCache);
var BlockLikeStorageCache =
/*#__PURE__*/
function (_BaseReadStorageCache2) {
  _inherits(BlockLikeStorageCache, _BaseReadStorageCache2);

  function BlockLikeStorageCache(options) {
    var _this13;

    _classCallCheck(this, BlockLikeStorageCache);
    _this13 = _possibleConstructorReturn(this, (BlockLikeStorageCache.__proto__ || Object.getPrototypeOf(BlockLikeStorageCache)).call(this, {
      readStorage: options.readStorage,
      name: options.name,
      createAddChange: options.createAddChange
    }));
    _this13._indexValues = {};
    return _this13;
  }

  _createClass(BlockLikeStorageCache, [{
    key: "add",
    value: function () {
      var _ref11 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee6(value, force) {
        var currentValue, addValue;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                if (force) {
                  _context6.next = 6;
                  break;
                }

                _context6.next = 3;
                return this.tryGet({
                  hashOrIndex: value.index
                });

              case 3:
                currentValue = _context6.sent;

                if (!(currentValue != null)) {
                  _context6.next = 6;
                  break;
                }

                throw new Error('Attempted to add an already existing object.');

              case 6:
                addValue = {
                  type: 'add',
                  value: value
                };
                this._values[neoBlockchainCore.common.uInt256ToString(value.hash)] = addValue;
                this._indexValues["".concat(value.index)] = addValue;

              case 9:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function add(_x8, _x9) {
        return _ref11.apply(this, arguments);
      }

      return add;
    }()
  }, {
    key: "_tryGetTracked",
    value: function _tryGetTracked(key) {
      if (typeof key.hashOrIndex !== 'number') {
        return this._values[neoBlockchainCore.common.uInt256ToString(key.hashOrIndex)];
      }

      return this._indexValues["".concat(key.hashOrIndex)];
    }
  }]);
  return BlockLikeStorageCache;
}(BaseReadStorageCache);

var WriteBatchBlockchain =
/*#__PURE__*/
function () {
  function WriteBatchBlockchain(options) {
    _classCallCheck(this, WriteBatchBlockchain);
    this.settings = options.settings;
    this._currentBlock = options.currentBlock;
    this._currentHeader = options.currentHeader;
    this._storage = options.storage;
    this._vm = options.vm;
    this._onStep = options.onStep;
    this._caches = {
      account: new ReadAllAddUpdateDeleteStorageCache({
        name: 'account',
        readAllStorage: this._storage.account,
        update: function update(value, _update) {
          return value.update(_update);
        },
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt160ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'account',
            value: value
          };
        },
        createDeleteChange: function createDeleteChange(key) {
          return {
            type: 'account',
            key: key
          };
        }
      }),
      action: new ReadGetAllAddStorageCache({
        name: 'action',
        readGetAllStorage: this._storage.action,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            blockIndex: value.blockIndex,
            transactionIndex: value.transactionIndex,
            index: value.index
          };
        },
        getKeyString: function getKeyString(key) {
          return "".concat(key.blockIndex, ":").concat(key.transactionIndex, ":").concat(key.index);
        },
        // eslint-disable-next-line
        matchesPartialKey: function matchesPartialKey(value, key) {
          // TODO: Implement me
          throw new Error('Not implemented');
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'action',
            value: value
          };
        }
      }),
      asset: new ReadAddUpdateStorageCache({
        name: 'asset',
        readStorage: this._storage.asset,
        update: function update(value, _update2) {
          return value.update(_update2);
        },
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt256ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'asset',
            value: value
          };
        }
      }),
      block: new BlockLikeStorageCache({
        name: 'block',
        readStorage: {
          get: this._storage.block.get,
          tryGet: this._storage.block.tryGet
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'block',
            value: value
          };
        }
      }),
      blockSystemFee: new ReadAddStorageCache({
        name: 'blockSystemFee',
        readStorage: this._storage.blockSystemFee,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt256ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'blockSystemFee',
            value: value
          };
        }
      }),
      header: new BlockLikeStorageCache({
        name: 'header',
        readStorage: {
          get: this._storage.header.get,
          tryGet: this._storage.header.tryGet
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'header',
            value: value
          };
        }
      }),
      transaction: new ReadAddStorageCache({
        name: 'transaction',
        readStorage: this._storage.transaction,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt256ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'transaction',
            value: value
          };
        }
      }),
      transactionSpentCoins: new ReadAddUpdateStorageCache({
        name: 'transactionSpentCoins',
        readStorage: this._storage.transactionSpentCoins,
        update: function update(value, _update3) {
          return value.update(_update3);
        },
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt256ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'transactionSpentCoins',
            value: value
          };
        }
      }),
      output: this._storage.output,
      contract: new ReadAddDeleteStorageCache({
        name: 'contract',
        readStorage: this._storage.contract,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt160ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'contract',
            value: value
          };
        },
        createDeleteChange: function createDeleteChange(key) {
          return {
            type: 'contract',
            key: key
          };
        }
      }),
      storageItem: new ReadGetAllAddUpdateDeleteStorageCache({
        name: 'storageItem',
        readGetAllStorage: this._storage.storageItem,
        update: function update(value, _update4) {
          return value.update(_update4);
        },
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash,
            key: value.key
          };
        },
        getKeyString: function getKeyString(key) {
          return "".concat(neoBlockchainCore.common.uInt160ToString(key.hash), ":").concat(key.key.toString('hex'));
        },
        matchesPartialKey: function matchesPartialKey(value, key) {
          return key.hash == null || neoBlockchainCore.common.uInt160Equal(value.hash, key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'storageItem',
            value: value
          };
        },
        createDeleteChange: function createDeleteChange(key) {
          return {
            type: 'storageItem',
            key: key
          };
        }
      }),
      validator: new ReadAllAddStorageCache({
        name: 'validator',
        readAllStorage: this._storage.validator,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            publicKey: value.publicKey
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.ecPointToString(key.publicKey);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'validator',
            value: value
          };
        }
      }),
      invocationData: new ReadAddStorageCache({
        name: 'invocationData',
        readStorage: this._storage.invocationData,
        getKeyFromValue: function getKeyFromValue(value) {
          return {
            hash: value.hash
          };
        },
        getKeyString: function getKeyString(key) {
          return neoBlockchainCore.common.uInt256ToString(key.hash);
        },
        createAddChange: function createAddChange(value) {
          return {
            type: 'invocationData',
            value: value
          };
        }
      })
    };
    this.account = this._caches.account;
    this.action = this._caches.action;
    this.asset = this._caches.asset;
    this.block = this._caches.block;
    this.blockSystemFee = this._caches.blockSystemFee;
    this.header = this._caches.header;
    this.transaction = this._caches.transaction;
    this.transactionSpentCoins = this._caches.transactionSpentCoins;
    this.output = this._caches.output;
    this.contract = this._caches.contract;
    this.storageItem = this._caches.storageItem;
    this.validator = this._caches.validator;
    this.invocationData = this._caches.invocationData;
  }

  _createClass(WriteBatchBlockchain, [{
    key: "getChangeSet",
    value: function getChangeSet() {
      return this.account.getChangeSet().concat(this.asset.getChangeSet()).concat(this.action.getChangeSet()).concat(this.block.getChangeSet()).concat(this.blockSystemFee.getChangeSet()).concat(this.header.getChangeSet()).concat(this.transaction.getChangeSet()).concat(this.transactionSpentCoins.getChangeSet()).concat(this.contract.getChangeSet()).concat(this.storageItem.getChangeSet()).concat(this.validator.getChangeSet()).concat(this.invocationData.getChangeSet());
    }
  }, {
    key: "persistBlock",
    value: function () {
      var _ref = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(block) {
        var _ref2, _ref3, systemFee, _$$1, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, _ref6, _ref5, _idx, _transaction2;

        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all([block.index === 0 ? Promise.resolve(neoBlockchainCore.utils.ZERO) : this.blockSystemFee.get({
                  hash: block.previousHash
                }).then(function (blockSystemFee) {
                  return blockSystemFee.systemFee;
                }), this.block.add(block), this.header.add(block.header)]);

              case 2:
                _ref2 = _context.sent;
                _ref3 = _slicedToArray(_ref2, 2);
                systemFee = _ref3[0];
                _$$1 = _ref3[1];
                _context.next = 8;
                return this.blockSystemFee.add(new neoBlockchainNodeCore.BlockSystemFee({
                  hash: block.hash,
                  systemFee: systemFee.add(block.getSystemFee({
                    getOutput: this.output.get,
                    governingToken: this.settings.governingToken,
                    utilityToken: this.settings.utilityToken,
                    fees: this.settings.fees
                  }))
                }));

              case 8:
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context.prev = 11;
                _iterator = block.transactions.entries()[Symbol.iterator]();

              case 13:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context.next = 23;
                  break;
                }

                _ref6 = _step.value;
                _ref5 = _slicedToArray(_ref6, 2);
                _idx = _ref5[0];
                _transaction2 = _ref5[1];
                _context.next = 20;
                return this._persistTransaction(block, _transaction2, _idx);

              case 20:
                _iteratorNormalCompletion = true;
                _context.next = 13;
                break;

              case 23:
                _context.next = 29;
                break;

              case 25:
                _context.prev = 25;
                _context.t0 = _context["catch"](11);
                _didIteratorError = true;
                _iteratorError = _context.t0;

              case 29:
                _context.prev = 29;
                _context.prev = 30;

                if (!_iteratorNormalCompletion && _iterator.return != null) {
                  _iterator.return();
                }

              case 32:
                _context.prev = 32;

                if (!_didIteratorError) {
                  _context.next = 35;
                  break;
                }

                throw _iteratorError;

              case 35:
                return _context.finish(32);

              case 36:
                return _context.finish(29);

              case 37:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this, [[11, 25, 29, 37], [30,, 32, 36]]);
      }));

      function persistBlock(_x) {
        return _ref.apply(this, arguments);
      }

      return persistBlock;
    }()
  }, {
    key: "_persistTransaction",
    value: function () {
      var _ref7 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(block, transaction, transactionIndex) {
        var _this = this;

        var results, _contract, temporaryBlockchain, result, assetChangeSet, assetHash, contractsChangeSet, contractHashes;

        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return Promise.all([this.transaction.add(transaction), this.transactionSpentCoins.add(new neoBlockchainNodeCore.TransactionSpentCoins({
                  hash: transaction.hash,
                  startHeight: block.index
                })), this._updateAccounts(transaction.inputs, transaction.outputs), this._updateCoins(transaction.inputs, transaction.type === neoBlockchainCore.TRANSACTION_TYPE.CLAIM && transaction instanceof neoBlockchainCore.ClaimTransaction ? transaction.claims : [], block)]);

              case 2:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.REGISTER && transaction instanceof neoBlockchainCore.RegisterTransaction)) {
                  _context4.next = 7;
                  break;
                }

                _context4.next = 5;
                return this.asset.add(new neoBlockchainCore.Asset({
                  hash: transaction.hash,
                  type: transaction.asset.type,
                  name: transaction.asset.name,
                  amount: transaction.asset.amount,
                  precision: transaction.asset.precision,
                  owner: transaction.asset.owner,
                  admin: transaction.asset.admin,
                  issuer: transaction.asset.admin,
                  expiration: this.currentBlockIndex + 2 * 2000000,
                  isFrozen: false
                }));

              case 5:
                _context4.next = 47;
                break;

              case 7:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.ISSUE && transaction instanceof neoBlockchainCore.IssueTransaction)) {
                  _context4.next = 15;
                  break;
                }

                _context4.next = 10;
                return neoBlockchainCore.utils.entries(transaction.getTransactionResults({
                  getOutput: this.output.get
                }));

              case 10:
                results = _context4.sent;
                _context4.next = 13;
                return Promise.all(results.map(function () {
                  var _ref9 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee2(_ref8) {
                    var _ref10, assetHex, value, hash, asset;

                    return regeneratorRuntime.wrap(function _callee2$(_context2) {
                      while (1) {
                        switch (_context2.prev = _context2.next) {
                          case 0:
                            _ref10 = _slicedToArray(_ref8, 2), assetHex = _ref10[0], value = _ref10[1];
                            hash = neoBlockchainCore.common.stringToUInt256(assetHex);
                            _context2.next = 4;
                            return _this.asset.get({
                              hash: hash
                            });

                          case 4:
                            asset = _context2.sent;
                            _context2.next = 7;
                            return _this.asset.update(asset, {
                              available: asset.available.add(value.neg())
                            });

                          case 7:
                          case "end":
                            return _context2.stop();
                        }
                      }
                    }, _callee2, this);
                  }));

                  return function (_x5) {
                    return _ref9.apply(this, arguments);
                  };
                }()));

              case 13:
                _context4.next = 47;
                break;

              case 15:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.ENROLLMENT && transaction instanceof neoBlockchainCore.EnrollmentTransaction)) {
                  _context4.next = 20;
                  break;
                }

                _context4.next = 18;
                return this.validator.add(new neoBlockchainCore.Validator({
                  publicKey: transaction.publicKey
                }));

              case 18:
                _context4.next = 47;
                break;

              case 20:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.PUBLISH && transaction instanceof neoBlockchainCore.PublishTransaction)) {
                  _context4.next = 29;
                  break;
                }

                _context4.next = 23;
                return this.contract.tryGet({
                  hash: transaction.contract.hash
                });

              case 23:
                _contract = _context4.sent;

                if (!(_contract == null)) {
                  _context4.next = 27;
                  break;
                }

                _context4.next = 27;
                return this.contract.add(transaction.contract);

              case 27:
                _context4.next = 47;
                break;

              case 29:
                if (!(transaction.type === neoBlockchainCore.TRANSACTION_TYPE.INVOCATION && transaction instanceof neoBlockchainCore.InvocationTransaction)) {
                  _context4.next = 47;
                  break;
                }

                _context4.prev = 30;
                // TODO: Make this less hacky.
                temporaryBlockchain = new WriteBatchBlockchain({
                  settings: this.settings,
                  currentBlock: this._currentBlock,
                  currentHeader: this._currentHeader,
                  storage: this,
                  vm: this._vm,
                  onStep: this._onStep
                });
                _context4.next = 34;
                return this._vm.executeScripts({
                  scripts: [{
                    code: transaction.script
                  }],
                  blockchain: temporaryBlockchain,
                  scriptContainer: {
                    type: neoBlockchainCore.SCRIPT_CONTAINER_TYPE.TRANSACTION,
                    value: transaction
                  },
                  triggerType: neoBlockchainNodeCore.TRIGGER_TYPE.APPLICATION,
                  action: {
                    blockIndex: block.index,
                    blockHash: block.hash,
                    transactionIndex: transactionIndex,
                    transactionHash: transaction.hash
                  },
                  gas: transaction.gas,
                  onStep: this._onStep
                });

              case 34:
                result = _context4.sent;
                assetChangeSet = temporaryBlockchain.asset.getChangeSet();
                assetHash = assetChangeSet.map(function (change) {
                  return change.type === 'add' && change.change.type === 'asset' ? change.change.value.hash : null;
                }).find(function (value) {
                  return value != null;
                });
                contractsChangeSet = temporaryBlockchain.contract.getChangeSet();
                contractHashes = contractsChangeSet.map(function (change) {
                  return change.type === 'add' && change.change.type === 'contract' ? change.change.value.hash : null;
                }).filter(Boolean);
                _context4.next = 41;
                return Promise.all([Promise.all(temporaryBlockchain.getChangeSet().map(function () {
                  var _ref11 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee3(change) {
                    return regeneratorRuntime.wrap(function _callee3$(_context3) {
                      while (1) {
                        switch (_context3.prev = _context3.next) {
                          case 0:
                            if (!(change.type === 'add')) {
                              _context3.next = 5;
                              break;
                            }

                            _context3.next = 3;
                            return _this._caches[change.change.type].add(change.change.value, true);

                          case 3:
                            _context3.next = 8;
                            break;

                          case 5:
                            if (!(change.type === 'delete')) {
                              _context3.next = 8;
                              break;
                            }

                            _context3.next = 8;
                            return _this._caches[change.change.type].delete(change.change.key);

                          case 8:
                          case "end":
                            return _context3.stop();
                        }
                      }
                    }, _callee3, this);
                  }));

                  return function (_x6) {
                    return _ref11.apply(this, arguments);
                  };
                }())), this.invocationData.add(new neoBlockchainCore.InvocationData({
                  hash: transaction.hash,
                  assetHash: assetHash,
                  contractHashes: contractHashes,
                  blockIndex: block.index,
                  transactionIndex: transactionIndex,
                  result: new neoBlockchainCore.InvocationResultSuccess({
                    stack: result.stack,
                    stackAlt: result.stackAlt
                  })
                }))]);

              case 41:
                _context4.next = 47;
                break;

              case 43:
                _context4.prev = 43;
                _context4.t0 = _context4["catch"](30);
                _context4.next = 47;
                return this.invocationData.add(new neoBlockchainCore.InvocationData({
                  hash: transaction.hash,
                  assetHash: undefined,
                  contractHashes: [],
                  blockIndex: block.index,
                  transactionIndex: transactionIndex,
                  result: new neoBlockchainCore.InvocationResultError({
                    message: _context4.t0.message
                  })
                }));

              case 47:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this, [[30, 43]]);
      }));

      function _persistTransaction(_x2, _x3, _x4) {
        return _ref7.apply(this, arguments);
      }

      return _persistTransaction;
    }()
  }, {
    key: "_updateAccounts",
    value: function () {
      var _ref12 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee5(inputs, outputs) {
        var _this2 = this;

        var inputOutputs, addressValues;
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return Promise.all(inputs.map(function (input) {
                  return _this2.output.get(input);
                }));

              case 2:
                inputOutputs = _context5.sent;
                addressValues = neoBlockchainCore.utils.entries(_.groupBy(inputOutputs.map(function (output) {
                  return [output.address, output.asset, output.value.neg()];
                }).concat(outputs.map(function (output) {
                  return [output.address, output.asset, output.value];
                })), // eslint-disable-next-line
                function (_ref13) {
                  var _ref14 = _slicedToArray(_ref13, 3),
                      address = _ref14[0],
                      asset = _ref14[1],
                      value = _ref14[2];

                  return neoBlockchainCore.common.uInt160ToHex(address);
                }));
                _context5.next = 6;
                return Promise.all(addressValues.map(function (_ref15) {
                  var _ref16 = _slicedToArray(_ref15, 2),
                      address = _ref16[0],
                      values = _ref16[1];

                  return _this2._updateAccount(neoBlockchainCore.common.hexToUInt160(address), // eslint-disable-next-line
                  values.map(function (_ref17) {
                    var _ref18 = _slicedToArray(_ref17, 3),
                        _$$1 = _ref18[0],
                        asset = _ref18[1],
                        value = _ref18[2];

                    return [asset, value];
                  }));
                }));

              case 6:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function _updateAccounts(_x7, _x8) {
        return _ref12.apply(this, arguments);
      }

      return _updateAccounts;
    }()
  }, {
    key: "_updateAccount",
    value: function () {
      var _ref19 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee6(address, values) {
        var account, balances, newAccount;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _context6.next = 2;
                return this.account.tryGet({
                  hash: address
                });

              case 2:
                account = _context6.sent;
                balances = values.reduce(function (acc, _ref20) {
                  var _ref21 = _slicedToArray(_ref20, 2),
                      asset = _ref21[0],
                      value = _ref21[1];

                  var key = neoBlockchainCore.common.uInt256ToHex(asset);

                  if (acc[key] == null) {
                    acc[key] = neoBlockchainCore.utils.ZERO;
                  }

                  acc[key] = acc[key].add(value);
                  return acc;
                }, account == null ? {} : account.balances);

                if (!(account == null)) {
                  _context6.next = 9;
                  break;
                }

                _context6.next = 7;
                return this.account.add(new neoBlockchainCore.Account({
                  hash: address,
                  balances: balances
                }));

              case 7:
                _context6.next = 15;
                break;

              case 9:
                _context6.next = 11;
                return this.account.update(account, {
                  balances: balances
                });

              case 11:
                newAccount = _context6.sent;

                if (!newAccount.isDeletable()) {
                  _context6.next = 15;
                  break;
                }

                _context6.next = 15;
                return this.account.delete({
                  hash: address
                });

              case 15:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function _updateAccount(_x9, _x10) {
        return _ref19.apply(this, arguments);
      }

      return _updateAccount;
    }()
  }, {
    key: "_updateCoins",
    value: function () {
      var _ref22 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee7(inputs, claims, block) {
        var _this3 = this;

        var inputClaims, hashInputClaims;
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                inputClaims = inputs.map(function (input) {
                  return {
                    type: 'input',
                    input: input,
                    hash: input.hash
                  };
                }).concat(claims.map(function (input) {
                  return {
                    type: 'claim',
                    input: input,
                    hash: input.hash
                  };
                }));
                hashInputClaims = neoBlockchainCore.utils.entries(_.groupBy(inputClaims, function (_ref23) {
                  var hash = _ref23.hash;
                  return neoBlockchainCore.common.uInt256ToHex(hash);
                }));
                _context7.next = 4;
                return Promise.all(hashInputClaims.map(function (_ref24) {
                  var _ref25 = _slicedToArray(_ref24, 2),
                      hash = _ref25[0],
                      values = _ref25[1];

                  return _this3._updateCoin(neoBlockchainCore.common.hexToUInt256(hash), values, block);
                }));

              case 4:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _updateCoins(_x11, _x12, _x13) {
        return _ref22.apply(this, arguments);
      }

      return _updateCoins;
    }()
  }, {
    key: "_updateCoin",
    value: function () {
      var _ref26 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee8(hash, inputClaims, block) {
        var spentCoins, endHeights, claimed, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _inputClaim;

        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                _context8.next = 2;
                return this.transactionSpentCoins.get({
                  hash: hash
                });

              case 2:
                spentCoins = _context8.sent;
                endHeights = _extends({}, spentCoins.endHeights);
                claimed = _extends({}, spentCoins.claimed);
                _iteratorNormalCompletion2 = true;
                _didIteratorError2 = false;
                _iteratorError2 = undefined;
                _context8.prev = 8;

                for (_iterator2 = inputClaims[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                  _inputClaim = _step2.value;

                  if (_inputClaim.type === 'input') {
                    endHeights[_inputClaim.input.index] = block.index;
                  } else {
                    claimed[_inputClaim.input.index] = true;
                  }
                }

                _context8.next = 16;
                break;

              case 12:
                _context8.prev = 12;
                _context8.t0 = _context8["catch"](8);
                _didIteratorError2 = true;
                _iteratorError2 = _context8.t0;

              case 16:
                _context8.prev = 16;
                _context8.prev = 17;

                if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
                  _iterator2.return();
                }

              case 19:
                _context8.prev = 19;

                if (!_didIteratorError2) {
                  _context8.next = 22;
                  break;
                }

                throw _iteratorError2;

              case 22:
                return _context8.finish(19);

              case 23:
                return _context8.finish(16);

              case 24:
                _context8.next = 26;
                return this.transactionSpentCoins.update(spentCoins, {
                  endHeights: endHeights,
                  claimed: claimed
                });

              case 26:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this, [[8, 12, 16, 24], [17,, 19, 23]]);
      }));

      function _updateCoin(_x14, _x15, _x16) {
        return _ref26.apply(this, arguments);
      }

      return _updateCoin;
    }()
  }, {
    key: "currentBlock",
    get: function get() {
      if (this._currentBlock == null) {
        throw new GenesisBlockNotRegisteredError();
      }

      return this._currentBlock;
    }
  }, {
    key: "currentBlockIndex",
    get: function get() {
      return this._currentBlock == null ? 0 : this._currentBlock.index;
    }
  }, {
    key: "currentHeader",
    get: function get() {
      if (this._currentHeader == null) {
        throw new GenesisBlockNotRegisteredError();
      }

      return this._currentHeader;
    }
  }]);
  return WriteBatchBlockchain;
}();

// $FlowFixMe

var Blockchain$1 =
/*#__PURE__*/
function () {
  function Blockchain(options) {
    var _this = this;

    _classCallCheck(this, Blockchain);
    Object.defineProperty(this, "_verifyScript", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref2 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee(_ref) {
          var scriptContainer, hash, witness, blockchain, result, stack, top;
          return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
              switch (_context.prev = _context.next) {
                case 0:
                  scriptContainer = _ref.scriptContainer, hash = _ref.hash, witness = _ref.witness;

                  if (neoBlockchainCore.common.uInt160Equal(hash, neoBlockchainCore.crypto.toScriptHash(witness.verification))) {
                    _context.next = 3;
                    break;
                  }

                  throw new WitnessVerifyError();

                case 3:
                  blockchain = _this._createWriteBlockchain();
                  _context.next = 6;
                  return _this._vm.executeScripts({
                    scripts: [{
                      code: witness.invocation,
                      pushOnly: true
                    }, {
                      code: witness.verification
                    }],
                    blockchain: blockchain,
                    scriptContainer: scriptContainer,
                    triggerType: neoBlockchainNodeCore.TRIGGER_TYPE.VERIFICATION,
                    action: neoBlockchainNodeCore.NULL_ACTION,
                    gas: neoBlockchainCore.utils.ZERO,
                    onStep: _this._onStep
                  });

                case 6:
                  result = _context.sent;

                  if (!(blockchain.getChangeSet().length !== 0)) {
                    _context.next = 9;
                    break;
                  }

                  throw new VerifyError();

                case 9:
                  stack = result.stack;

                  if (!(stack.length !== 1)) {
                    _context.next = 12;
                    break;
                  }

                  throw new VerifyError();

                case 12:
                  top = stack[0];

                  if (top.asBoolean()) {
                    _context.next = 15;
                    break;
                  }

                  throw new VerifyError();

                case 15:
                case "end":
                  return _context.stop();
              }
            }
          }, _callee, this);
        }));

        function value(_x) {
          return _ref2.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "calculateClaimAmount", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref3 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee3(claims) {
          var spentCoins, filteredSpentCoins;
          return regeneratorRuntime.wrap(function _callee3$(_context3) {
            while (1) {
              switch (_context3.prev = _context3.next) {
                case 0:
                  _context3.next = 2;
                  return Promise.all(claims.map(function (claim) {
                    return _this._tryGetSpentCoin(claim);
                  }));

                case 2:
                  spentCoins = _context3.sent;
                  filteredSpentCoins = spentCoins.filter(Boolean);

                  if (!(spentCoins.length !== filteredSpentCoins.length)) {
                    _context3.next = 6;
                    break;
                  }

                  throw new Error('Not all coins were spent');

                case 6:
                  if (!filteredSpentCoins.some(function (coin) {
                    return coin.claimed;
                  })) {
                    _context3.next = 8;
                    break;
                  }

                  throw new Error('Coin was already claimed');

                case 8:
                  if (!filteredSpentCoins.some(function (coin) {
                    return !neoBlockchainCore.common.uInt256Equal(coin.output.asset, _this.settings.governingToken.hash);
                  })) {
                    _context3.next = 10;
                    break;
                  }

                  throw new Error('Invalid claim');

                case 10:
                  return _context3.abrupt("return", neoBlockchainCore.utils.calculateClaimAmount({
                    coins: filteredSpentCoins.map(function (coin) {
                      return {
                        value: coin.output.value,
                        startHeight: coin.startHeight,
                        endHeight: coin.endHeight
                      };
                    }),
                    decrementInterval: _this.settings.decrementInterval,
                    generationAmount: _this.settings.generationAmount,
                    getSystemFee: function () {
                      var _ref4 = _asyncToGenerator(
                      /*#__PURE__*/
                      regeneratorRuntime.mark(function _callee2(index) {
                        var header, blockSystemFee;
                        return regeneratorRuntime.wrap(function _callee2$(_context2) {
                          while (1) {
                            switch (_context2.prev = _context2.next) {
                              case 0:
                                _context2.next = 2;
                                return _this._storage.header.get({
                                  hashOrIndex: index
                                });

                              case 2:
                                header = _context2.sent;
                                _context2.next = 5;
                                return _this._storage.blockSystemFee.get({
                                  hash: header.hash
                                });

                              case 5:
                                blockSystemFee = _context2.sent;
                                return _context2.abrupt("return", blockSystemFee.systemFee);

                              case 7:
                              case "end":
                                return _context2.stop();
                            }
                          }
                        }, _callee2, this);
                      }));

                      function getSystemFee(_x3) {
                        return _ref4.apply(this, arguments);
                      }

                      return getSystemFee;
                    }()
                  }));

                case 11:
                case "end":
                  return _context3.stop();
              }
            }
          }, _callee3, this);
        }));

        function value(_x2) {
          return _ref3.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_isSpent", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref5 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee4(input) {
          var transactionSpentCoins;
          return regeneratorRuntime.wrap(function _callee4$(_context4) {
            while (1) {
              switch (_context4.prev = _context4.next) {
                case 0:
                  _context4.next = 2;
                  return _this.transactionSpentCoins.tryGet({
                    hash: input.hash
                  });

                case 2:
                  transactionSpentCoins = _context4.sent;
                  return _context4.abrupt("return", transactionSpentCoins != null && transactionSpentCoins.endHeights[input.index] != null);

                case 4:
                case "end":
                  return _context4.stop();
              }
            }
          }, _callee4, this);
        }));

        function value(_x4) {
          return _ref5.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_tryGetSpentCoin", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref6 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee5(input) {
          var _ref7, _ref8, transactionSpentCoins, output, endHeight, claimed;

          return regeneratorRuntime.wrap(function _callee5$(_context5) {
            while (1) {
              switch (_context5.prev = _context5.next) {
                case 0:
                  _context5.next = 2;
                  return Promise.all([_this.transactionSpentCoins.tryGet({
                    hash: input.hash
                  }), _this.output.get(input)]);

                case 2:
                  _ref7 = _context5.sent;
                  _ref8 = _slicedToArray(_ref7, 2);
                  transactionSpentCoins = _ref8[0];
                  output = _ref8[1];

                  if (!(transactionSpentCoins == null)) {
                    _context5.next = 8;
                    break;
                  }

                  return _context5.abrupt("return", null);

                case 8:
                  endHeight = transactionSpentCoins.endHeights[input.index];

                  if (!(endHeight == null)) {
                    _context5.next = 11;
                    break;
                  }

                  return _context5.abrupt("return", null);

                case 11:
                  claimed = transactionSpentCoins.claimed[input.index];
                  return _context5.abrupt("return", {
                    output: output,
                    startHeight: transactionSpentCoins.startHeight,
                    endHeight: endHeight,
                    claimed: !!claimed
                  });

                case 13:
                case "end":
                  return _context5.stop();
              }
            }
          }, _callee5, this);
        }));

        function value(_x5) {
          return _ref6.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_getValidators", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref9 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee6(transactions) {
          var _ref10, _ref11, votes, validators, sortedVotes, validatorsCount, validatorsToCount, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, _vote, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _publicKey, publicKeyHex;

          return regeneratorRuntime.wrap(function _callee6$(_context6) {
            while (1) {
              switch (_context6.prev = _context6.next) {
                case 0:
                  _context6.next = 2;
                  return Promise.all([_this._getVotes(transactions), _this.validator.all.toArray().toPromise()]);

                case 2:
                  _ref10 = _context6.sent;
                  _ref11 = _slicedToArray(_ref10, 2);
                  votes = _ref11[0];
                  validators = _ref11[1];
                  sortedVotes = _.sortBy(votes, function (vote) {
                    return vote.publicKeys.length;
                  });
                  validatorsCount = Math.max(neoBlockchainCore.utils.weightedAverage(neoBlockchainCore.utils.weightedFilter(sortedVotes, 0.25, 0.75, function (vote) {
                    return vote.count.toNumber();
                  }).map(function (_ref12) {
                    var _ref13 = _slicedToArray(_ref12, 2),
                        vote = _ref13[0],
                        weight = _ref13[1];

                    return {
                      value: vote.publicKeys.length,
                      weight: weight
                    };
                  })), _this.settings.standbyValidators.length);
                  validatorsToCount = _.fromPairs(validators.map(function (validator) {
                    return [neoBlockchainCore.common.ecPointToHex(validator.publicKey), neoBlockchainCore.utils.ZERO];
                  }));
                  _iteratorNormalCompletion = true;
                  _didIteratorError = false;
                  _iteratorError = undefined;
                  _context6.prev = 12;
                  _iterator = votes[Symbol.iterator]();

                case 14:
                  if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                    _context6.next = 38;
                    break;
                  }

                  _vote = _step.value;
                  _iteratorNormalCompletion2 = true;
                  _didIteratorError2 = false;
                  _iteratorError2 = undefined;
                  _context6.prev = 19;

                  for (_iterator2 = _.take(_vote.publicKeys, validatorsCount)[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                    _publicKey = _step2.value;
                    publicKeyHex = neoBlockchainCore.common.ecPointToHex(_publicKey);

                    if (validatorsToCount[publicKeyHex] != null) {
                      validatorsToCount[publicKeyHex] = validatorsToCount[publicKeyHex].add(_vote.count);
                    }
                  }

                  _context6.next = 27;
                  break;

                case 23:
                  _context6.prev = 23;
                  _context6.t0 = _context6["catch"](19);
                  _didIteratorError2 = true;
                  _iteratorError2 = _context6.t0;

                case 27:
                  _context6.prev = 27;
                  _context6.prev = 28;

                  if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
                    _iterator2.return();
                  }

                case 30:
                  _context6.prev = 30;

                  if (!_didIteratorError2) {
                    _context6.next = 33;
                    break;
                  }

                  throw _iteratorError2;

                case 33:
                  return _context6.finish(30);

                case 34:
                  return _context6.finish(27);

                case 35:
                  _iteratorNormalCompletion = true;
                  _context6.next = 14;
                  break;

                case 38:
                  _context6.next = 44;
                  break;

                case 40:
                  _context6.prev = 40;
                  _context6.t1 = _context6["catch"](12);
                  _didIteratorError = true;
                  _iteratorError = _context6.t1;

                case 44:
                  _context6.prev = 44;
                  _context6.prev = 45;

                  if (!_iteratorNormalCompletion && _iterator.return != null) {
                    _iterator.return();
                  }

                case 47:
                  _context6.prev = 47;

                  if (!_didIteratorError) {
                    _context6.next = 50;
                    break;
                  }

                  throw _iteratorError;

                case 50:
                  return _context6.finish(47);

                case 51:
                  return _context6.finish(44);

                case 52:
                  return _context6.abrupt("return", _.take(neoBlockchainCore.utils.entries(validatorsToCount).sort(function (_ref14, _ref15) {
                    var _ref16 = _slicedToArray(_ref14, 2),
                        aKey = _ref16[0],
                        aValue = _ref16[1];

                    var _ref17 = _slicedToArray(_ref15, 2),
                        bKey = _ref17[0],
                        bValue = _ref17[1];

                    return aValue.eq(bValue) ? neoBlockchainCore.common.ecPointCompare(aKey, bKey) : -aValue.cmp(bValue);
                  }) // eslint-disable-next-line
                  .map(function (_ref18) {
                    var _ref19 = _slicedToArray(_ref18, 2),
                        key = _ref19[0],
                        _$$1 = _ref19[1];

                    return neoBlockchainCore.common.hexToECPoint(key);
                  }), validatorsCount));

                case 53:
                case "end":
                  return _context6.stop();
              }
            }
          }, _callee6, this, [[12, 40, 44, 52], [19, 23, 27, 35], [28,, 30, 34], [45,, 47, 51]]);
        }));

        function value(_x6) {
          return _ref9.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_getVotes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref20 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee7(transactions) {
          var inputs, outputs, changes, votes;
          return regeneratorRuntime.wrap(function _callee7$(_context7) {
            while (1) {
              switch (_context7.prev = _context7.next) {
                case 0:
                  _context7.next = 2;
                  return Promise.all(transactions.map(function (transaction) {
                    return transaction.getReferences({
                      getOutput: _this.output.get
                    });
                  })).then(function (results) {
                    return results.reduce(function (acc, inputResults) {
                      return acc.concat(inputResults);
                    }, []).map(function (output) {
                      return {
                        address: output.address,
                        asset: output.asset,
                        value: output.value.neg()
                      };
                    });
                  });

                case 2:
                  inputs = _context7.sent;
                  outputs = transactions.reduce(function (acc, transaction) {
                    return acc.concat(transaction.outputs);
                  }, []).map(function (output) {
                    return {
                      address: output.address,
                      asset: output.asset,
                      value: output.value
                    };
                  });
                  changes = _.fromPairs(neoBlockchainCore.utils.entries(_.groupBy(inputs.concat(outputs).filter(function (output) {
                    return neoBlockchainCore.common.uInt256Equal(output.asset, _this.settings.governingToken.hash);
                  }), function (output) {
                    return neoBlockchainCore.common.uInt160ToHex(output.address);
                  } // eslint-disable-next-line
                  )).map(function (_ref21) {
                    var _ref22 = _slicedToArray(_ref21, 2),
                        addressHex = _ref22[0],
                        addressOutputs = _ref22[1];

                    return [addressHex, addressOutputs.reduce(function (acc, output) {
                      return acc.add(output.value);
                    }, neoBlockchainCore.utils.ZERO)];
                  }));
                  _context7.next = 7;
                  return _this.account.all.filter(function (account) {
                    return account.votes.length > 0;
                  }).map(function (account) {
                    var balance = account.balances[_this.settings.governingToken.hashHex] || neoBlockchainCore.utils.ZERO;
                    balance = balance.add(changes[account.hashHex] || neoBlockchainCore.utils.ZERO);
                    return balance.lte(neoBlockchainCore.utils.ZERO) ? null : {
                      publicKeys: account.votes,
                      count: balance
                    };
                  }).toArray().toPromise();

                case 7:
                  votes = _context7.sent;

                  if (!(votes.length === 0)) {
                    _context7.next = 10;
                    break;
                  }

                  return _context7.abrupt("return", [{
                    publicKeys: _this.settings.standbyValidators,
                    count: _this.settings.governingToken.asset.amount
                  }]);

                case 10:
                  return _context7.abrupt("return", votes.filter(Boolean));

                case 11:
                case "end":
                  return _context7.stop();
              }
            }
          }, _callee7, this);
        }));

        function value(_x7) {
          return _ref20.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_getInvocationData", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function () {
        var _ref23 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee8(transaction) {
          var data, _ref24, _ref25, asset, contracts, actions;

          return regeneratorRuntime.wrap(function _callee8$(_context8) {
            while (1) {
              switch (_context8.prev = _context8.next) {
                case 0:
                  _context8.next = 2;
                  return _this._storage.invocationData.get({
                    hash: transaction.hash
                  });

                case 2:
                  data = _context8.sent;
                  _context8.next = 5;
                  return Promise.all([data.assetHash == null ? Promise.resolve(null) : _this._storage.asset.get({
                    hash: data.assetHash
                  }), Promise.all(data.contractHashes.map(function (contractHash) {
                    return _this._storage.contract.get({
                      hash: contractHash
                    });
                  })), // TODO: Make this more efficient with a DataLoader
                  _this._storage.action.getAll({
                    blockIndexStart: data.blockIndex,
                    transactionIndexStart: data.transactionIndex,
                    blockIndexStop: data.blockIndex,
                    transactionIndexStop: data.transactionIndex
                  }).toArray().toPromise()]);

                case 5:
                  _ref24 = _context8.sent;
                  _ref25 = _slicedToArray(_ref24, 3);
                  asset = _ref25[0];
                  contracts = _ref25[1];
                  actions = _ref25[2];
                  return _context8.abrupt("return", {
                    asset: asset,
                    contracts: contracts,
                    actions: actions,
                    result: data.result
                  });

                case 11:
                case "end":
                  return _context8.stop();
              }
            }
          }, _callee8, this);
        }));

        function value(_x8) {
          return _ref23.apply(this, arguments);
        }

        return value;
      }()
    });
    Object.defineProperty(this, "_onStep", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: function value(_ref26) {
        var context = _ref26.context,
            opCode = _ref26.opCode;
        var scriptHash = neoBlockchainCore.common.uInt160ToString(context.scriptHash);

        _this.log({
          event: 'VM_STEP',
          level: 'silly',
          scriptHash: scriptHash,
          pc: context.pc,
          opCode: opCode
        });
      }
    });
    this._storage = options.storage;
    this._currentBlock = options.currentBlock;
    this._currentHeader = options.currentHeader;
    this._persistingBlocks = false;
    this._blockQueue = new PriorityQueue({
      comparator: function comparator(a, b) {
        return a.block.index - b.block.index;
      }
    });
    this._inQueue = new Set();
    this._vm = options.vm;
    this._running = true;
    this._doneRunningResolve = null;
    this.settings = options.settings;
    this.log = options.log;
    this.account = this._storage.account;
    this.action = this._storage.action;
    this.asset = this._storage.asset;
    this.block = {
      get: this._storage.block.get,
      tryGet: this._storage.block.tryGet
    };
    this.blockSystemFee = this._storage.blockSystemFee;
    this.header = {
      get: this._storage.header.get,
      tryGet: this._storage.header.tryGet
    };
    this.transaction = this._storage.transaction;
    this.transactionSpentCoins = this._storage.transactionSpentCoins;
    this.output = this._storage.output;
    this.contract = this._storage.contract;
    this.storageItem = this._storage.storageItem;
    this.validator = this._storage.validator;
    this.invocationData = this._storage.invocationData;
    this.deserializeWireContext = {
      messageMagic: this.settings.messageMagic
    };
    this.serializeJSONContext = {
      addressVersion: this.settings.addressVersion,
      feeContext: {
        getOutput: this.output.get,
        governingToken: this.settings.governingToken,
        utilityToken: this.settings.utilityToken,
        fees: this.settings.fees
      },
      getInvocationData: this._getInvocationData
    };
  }

  _createClass(Blockchain, [{
    key: "stop",
    value: function () {
      var _ref27 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee9() {
        var _this2 = this;

        var doneRunningPromise;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                if (this._running) {
                  _context9.next = 2;
                  break;
                }

                return _context9.abrupt("return");

              case 2:
                if (!this._persistingBlocks) {
                  _context9.next = 9;
                  break;
                }

                doneRunningPromise = new Promise(function (resolve) {
                  _this2._doneRunningResolve = resolve;
                });
                this._running = false;
                _context9.next = 7;
                return doneRunningPromise;

              case 7:
                _context9.next = 10;
                break;

              case 9:
                this._running = false;

              case 10:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function stop() {
        return _ref27.apply(this, arguments);
      }

      return stop;
    }()
  }, {
    key: "persistBlock",
    value: function persistBlock(_ref28) {
      var _this3 = this;

      var block = _ref28.block,
          unsafe = _ref28.unsafe;
      return new Promise(function (resolve, reject) {
        if (_this3._inQueue.has(block.index)) {
          return;
        }

        _this3._inQueue.add(block.index);

        _this3._blockQueue.queue({
          block: block,
          resolve: resolve,
          reject: reject,
          unsafe: unsafe
        });

        _this3._persistBlocksAsync();
      });
    } // eslint-disable-next-line

  }, {
    key: "persistHeaders",
    value: function () {
      var _ref29 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee10(headers) {
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function persistHeaders(_x9) {
        return _ref29.apply(this, arguments);
      }

      return persistHeaders;
    }()
  }, {
    key: "verifyBlock",
    value: function () {
      var _ref30 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee11(block) {
        return regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                _context11.next = 2;
                return block.verify({
                  genesisBlock: this.settings.genesisBlock,
                  tryGetBlock: this.block.tryGet,
                  tryGetHeader: this.header.tryGet,
                  isSpent: this._isSpent,
                  getAsset: this.asset.get,
                  getOutput: this.output.get,
                  getValidators: this._getValidators,
                  calculateClaimAmount: this.calculateClaimAmount,
                  verifyScript: this._verifyScript,
                  currentHeight: this._currentBlock == null ? 0 : this._currentBlock.index,
                  governingToken: this.settings.governingToken,
                  utilityToken: this.settings.utilityToken,
                  fees: this.settings.fees
                });

              case 2:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      function verifyBlock(_x10) {
        return _ref30.apply(this, arguments);
      }

      return verifyBlock;
    }()
  }, {
    key: "verifyTransaction",
    value: function () {
      var _ref32 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee12(_ref31) {
        var transaction, memPool;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                transaction = _ref31.transaction, memPool = _ref31.memPool;
                _context12.next = 3;
                return transaction.verify({
                  calculateClaimAmount: this.calculateClaimAmount,
                  isSpent: this._isSpent,
                  getAsset: this.asset.get,
                  getOutput: this.output.get,
                  verifyScript: this._verifyScript,
                  governingToken: this.settings.governingToken,
                  utilityToken: this.settings.utilityToken,
                  fees: this.settings.fees,
                  currentHeight: this.currentBlockIndex,
                  memPool: memPool
                });

              case 3:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      function verifyTransaction(_x11) {
        return _ref32.apply(this, arguments);
      }

      return verifyTransaction;
    }()
  }, {
    key: "invokeScript",
    value: function () {
      var _ref33 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee13(script) {
        var blockchain, transaction, result;
        return regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                blockchain = this._createWriteBlockchain();
                transaction = new neoBlockchainCore.InvocationTransaction({
                  script: script,
                  gas: neoBlockchainCore.common.ONE_HUNDRED_FIXED8
                });
                _context13.prev = 2;
                _context13.next = 5;
                return this._vm.executeScripts({
                  scripts: [{
                    code: script
                  }],
                  blockchain: blockchain,
                  scriptContainer: {
                    type: neoBlockchainCore.SCRIPT_CONTAINER_TYPE.TRANSACTION,
                    value: transaction
                  },
                  triggerType: neoBlockchainNodeCore.TRIGGER_TYPE.APPLICATION,
                  action: neoBlockchainNodeCore.NULL_ACTION,
                  gas: transaction.gas,
                  onStep: this._onStep
                });

              case 5:
                result = _context13.sent;
                return _context13.abrupt("return", new neoBlockchainCore.InvocationResultSuccess({
                  stack: result.stack,
                  stackAlt: result.stackAlt
                }));

              case 9:
                _context13.prev = 9;
                _context13.t0 = _context13["catch"](2);
                return _context13.abrupt("return", new neoBlockchainCore.InvocationResultError({
                  message: _context13.t0.message
                }));

              case 12:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13, this, [[2, 9]]);
      }));

      function invokeScript(_x12) {
        return _ref33.apply(this, arguments);
      }

      return invokeScript;
    }()
  }, {
    key: "_persistBlocksAsync",
    value: function () {
      var _ref34 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee14() {
        var entry, start, duration;
        return regeneratorRuntime.wrap(function _callee14$(_context14) {
          while (1) {
            switch (_context14.prev = _context14.next) {
              case 0:
                if (!(this._persistingBlocks || !this._running)) {
                  _context14.next = 2;
                  break;
                }

                return _context14.abrupt("return");

              case 2:
                this._persistingBlocks = true;
                _context14.prev = 3;
                this.cleanBlockQueue();
                entry = this.peekBlockQueue();

              case 6:
                if (!(this._running && entry != null && entry.block.index === this.currentBlockIndex + 1)) {
                  _context14.next = 18;
                  break;
                }

                entry = this._blockQueue.dequeue();
                start = perf_hooks.performance.now(); // eslint-disable-next-line

                _context14.next = 11;
                return this._persistBlock(entry.block, entry.unsafe);

              case 11:
                entry.resolve();
                duration = perf_hooks.performance.now() - start;
                this.log({
                  event: 'PERSIST_BLOCK_SUCCESS',
                  index: entry.block.index,
                  durationMS: duration
                });
                this.cleanBlockQueue();
                entry = this.peekBlockQueue();
                _context14.next = 6;
                break;

              case 18:
                _context14.next = 24;
                break;

              case 20:
                _context14.prev = 20;
                _context14.t0 = _context14["catch"](3);

                if (entry != null) {
                  entry.reject(_context14.t0);
                }

                this.log({
                  event: 'PERSIST_BLOCK_ERROR',
                  error: _context14.t0,
                  index: entry == null ? null : entry.block.index
                });

              case 24:
                _context14.prev = 24;
                this._persistingBlocks = false;

                if (this._doneRunningResolve != null) {
                  this._doneRunningResolve();

                  this._doneRunningResolve = null;
                }

                return _context14.finish(24);

              case 28:
              case "end":
                return _context14.stop();
            }
          }
        }, _callee14, this, [[3, 20, 24, 28]]);
      }));

      function _persistBlocksAsync() {
        return _ref34.apply(this, arguments);
      }

      return _persistBlocksAsync;
    }()
  }, {
    key: "cleanBlockQueue",
    value: function cleanBlockQueue() {
      var entry = this.peekBlockQueue();

      while (entry != null && entry.block.index <= this.currentBlockIndex) {
        this._blockQueue.dequeue();

        entry.resolve();
        entry = this.peekBlockQueue();
      }
    }
  }, {
    key: "peekBlockQueue",
    value: function peekBlockQueue() {
      if (this._blockQueue.length > 0) {
        return this._blockQueue.peek();
      }

      return null;
    }
  }, {
    key: "_persistBlock",
    value: function () {
      var _ref35 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee15(block, unsafe) {
        var blockchain;
        return regeneratorRuntime.wrap(function _callee15$(_context15) {
          while (1) {
            switch (_context15.prev = _context15.next) {
              case 0:
                

                blockchain = this._createWriteBlockchain();
                _context15.next = 4;
                return blockchain.persistBlock(block);

              case 4:
                _context15.next = 6;
                return this._storage.commit(blockchain.getChangeSet());

              case 6:
                this._currentBlock = block;
                this._currentHeader = block.header;

              case 8:
              case "end":
                return _context15.stop();
            }
          }
        }, _callee15, this);
      }));

      function _persistBlock(_x13, _x14) {
        return _ref35.apply(this, arguments);
      }

      return _persistBlock;
    }()
  }, {
    key: "_createWriteBlockchain",
    value: function _createWriteBlockchain() {
      return new WriteBatchBlockchain({
        settings: this.settings,
        currentBlock: this._currentBlock,
        currentHeader: this._currentHeader,
        storage: this._storage,
        vm: this._vm,
        onStep: this._onStep
      });
    }
  }, {
    key: "currentBlock",
    get: function get() {
      if (this._currentBlock == null) {
        throw new GenesisBlockNotRegisteredError();
      }

      return this._currentBlock;
    }
  }, {
    key: "currentHeader",
    get: function get() {
      if (this._currentHeader == null) {
        throw new GenesisBlockNotRegisteredError();
      }

      return this._currentHeader;
    }
  }, {
    key: "currentBlockIndex",
    get: function get() {
      return this._currentBlock == null ? -1 : this._currentBlock.index;
    }
  }], [{
    key: "create",
    value: function () {
      var _ref37 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee16(_ref36) {
        var settings, storage, vm, log, _ref38, _ref39, currentBlock, currentHeader, blockchain;

        return regeneratorRuntime.wrap(function _callee16$(_context16) {
          while (1) {
            switch (_context16.prev = _context16.next) {
              case 0:
                settings = _ref36.settings, storage = _ref36.storage, vm = _ref36.vm, log = _ref36.log;
                _context16.next = 3;
                return Promise.all([storage.block.tryGetLatest(), storage.header.tryGetLatest()]);

              case 3:
                _ref38 = _context16.sent;
                _ref39 = _slicedToArray(_ref38, 2);
                currentBlock = _ref39[0];
                currentHeader = _ref39[1];
                blockchain = new Blockchain({
                  currentBlock: currentBlock,
                  currentHeader: currentHeader,
                  settings: settings,
                  storage: storage,
                  vm: vm,
                  log: log
                });

                if (!(currentHeader == null)) {
                  _context16.next = 11;
                  break;
                }

                _context16.next = 11;
                return blockchain.persistHeaders([settings.genesisBlock.header]);

              case 11:
                if (!(currentBlock == null)) {
                  _context16.next = 14;
                  break;
                }

                _context16.next = 14;
                return blockchain.persistBlock({
                  block: settings.genesisBlock
                });

              case 14:
                return _context16.abrupt("return", blockchain);

              case 15:
              case "end":
                return _context16.stop();
            }
          }
        }, _callee16, this);
      }));

      function create(_x15) {
        return _ref37.apply(this, arguments);
      }

      return create;
    }()
  }]);
  return Blockchain;
}();

module.exports = Blockchain$1;
//# sourceMappingURL=index.js.map
